/*   ____  __  _ _____ ____     _ _            _
**  / ___||  \/ |_   _|  _ \___| (_) ___ _ __ | |_
**  \___ \| |\/| || | | |_)/ __| | |/ _ \ '_ \| __|
**   ___) | |  | || | |  _| (__| | |  __/ | | | |_
**  |____/|_|  |_||_| |_|  \___|_|_|\___|_| |_|\__|
**
**  SMTPclient -- simple SMTP client
**
**  This program is a minimal SMTP client that takes an email
**  message body and passes it on to a SMTP server (default is the
**  MTA on the local host). Since it is completely self-supporting,
**  it is especially suitable for use in restricted environments.
**
**  ======================================================================
**
**  Copyright (c) 1997 Ralf S. Engelschall, All rights reserved.
**  Copyright (c) 2000, 2001 Davide Libenzi <davidel@xmailserver.org>, All rights reserved.
**
**  This program is free software; it may be redistributed and/or modified
**  only under the terms of either the Artistic License or the GNU General
**  Public License, which may be found in the SMTP source distribution.
**  Look at the file COPYING.
**
**  This program is distributed in the hope that it will be useful, but
**  WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  ======================================================================
**
**  smtpclient_main.c -- program source
**
**  Based on smtp.c as of August 11, 1995 from
**      W.Z. Venema,
**      Eindhoven University of Technology,
**      Department of Mathematics and Computer Science,
**      Den Dolech 2, P.O. Box 513, 5600 MB Eindhoven, The Netherlands.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>

#include "smtpclient_getopt.h"
#include "smtpclient_errno.h"
#include "smtpclient_vers.h"
#include "base64.h"




#define MAX_FATTACH         64

#define dprintf  if (verbose) printf
#define dvprintf if (verbose) vprintf




static char     host_name[256] = "";
static char     cc_addr[1024] = "";
static char     err_addr[1024] = "";
static char     from_addr[1024] = "";
static char     mailhost[1024] = "";
static int      mailport = 25;
static char     reply_addr[1024] = "";
static char     subject[1024] = "";
static int      mime_style = 0;
static int      verbose = 0;

static int      sock;
static int      ibytes = 0;
static int      cbyte = 0;
static char     skibuf[2048];
static int      obytes = 0;
static char     skobuf[2048];





/*
**  logging support
*/
void            logmsg(char *str,...)
{

    va_list         ap;
    char            buf[2048];

    va_start(ap, str);
    vsprintf(buf, str, ap);
    va_end(ap);

    fprintf(stderr, "SMTPclient: %s\n", buf);

}



int             socksend(int s, char *buf, int size)
{

    while (size > 0)
    {
        int             sndsz = send(s, buf, size, 0);


        if (sndsz < 0)
            return 0;

        size -= sndsz;
        buf += sndsz;
    }

    return 1;

}



int             sockflush(void)
{

    int             pendbytes = obytes;

    if (obytes > 0)
    {
        socksend(sock, skobuf, obytes);

        obytes = 0;
    }

    return pendbytes;

}




int             sockputc(int ch)
{

    if (obytes >= sizeof(skobuf))
    {
        socksend(sock, skobuf, obytes);

        obytes = 0;
    }

    skobuf[obytes++] = ch;

    if (ch == '\n')
    {
        socksend(sock, skobuf, obytes);

        obytes = 0;
    }

    return ch;

}



void            sockprintf(char *str,...)
{

    int             ii;
    va_list         ap;
    char            buf[2048];

    va_start(ap, str);
    vsprintf(buf, str, ap);
    va_end(ap);

    for (ii = 0; buf[ii] != 0; ii++)
        sockputc(buf[ii]);

}


int             sockgetc(void)
{

    if (ibytes == 0)
    {
        if ((ibytes = recv(sock, skibuf, sizeof(skibuf), 0)) <= 0)
            return EOF;

        cbyte = 0;
    }

    --ibytes;

    return (int) skibuf[cbyte++];

}



char           *sockgets(char *buf, int size)
{

    int             ii = 0;

    for (; ii < size;)
    {
        int             ch = sockgetc();

        if (ch == EOF)
        {
            if (ii == 0)
                return (NULL);
            break;
        }

        if (ch == '\r')
        {
            sockgetc();

            break;
        }
        else
            buf[ii++] = ch;
    }

    buf[ii] = 0;

    return buf;

}




/*
**  usage page
*/
void            usage(char const * prgname)
{

    fprintf(stderr, "Usage: %s [options] recipients ...\n", prgname);
    fprintf(stderr, "\n");
    fprintf(stderr, "Message Header Options:\n");
    fprintf(stderr, "  -s, --subject=STR      subject line of message\n");
    fprintf(stderr, "  -f, --from=ADDR        address of the sender\n");
    fprintf(stderr, "  -r, --reply-to=ADDR    address of the sender for replies\n");
    fprintf(stderr, "  -e, --errors-to=ADDR   address to send delivery errors to\n");
    fprintf(stderr, "  -c, --carbon-copy=ADDR address to send copy of message to\n");
    fprintf(stderr, "  -b, --body=FILE        message body file\n");
    fprintf(stderr, "  -a, --attach=FILE      attach file\n");
    fprintf(stderr, "  -D, --delete           delete body file on exit\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Processing Options:\n");
    fprintf(stderr, "  -S, --smtp-host=HOST   host where MTA can be contacted via SMTP\n");
    fprintf(stderr, "  -P, --smtp-port=NUM    port where MTA can be contacted via SMTP\n");
    fprintf(stderr, "  -M, --mime-encode      use MIME-style translation to quoted-printable\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Giving Feedback:\n");
    fprintf(stderr, "  -v, --verbose          enable verbose logging messages\n");
    fprintf(stderr, "  -V, --version          display version string\n");
    fprintf(stderr, "  -h, --help             display this page\n");
    fprintf(stderr, "\n");

}

/*
**  version page
*/
void            version(void)
{

    fprintf(stdout, "%s\n", SMTPclient_Hello);
    fprintf(stdout, "\n");
    fprintf(stdout, "Copyright (c) 1997 Ralf S. Engelschall, All rights reserved.\n");
    fprintf(stdout, "Copyright (c) 2000, 2001 Davide Libenzi <davidel@xmailserver.org>, All rights reserved.\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "This program is distributed in the hope that it will be useful,\n");
    fprintf(stdout, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    fprintf(stdout, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    fprintf(stdout, "the GNU General Public License for more details.\n");
    fprintf(stdout, "\n");

}


/*
**  examine message from server
*/
void            get_response(void)
{

    char            buf[2048];

    while (sockgets(buf, sizeof(buf) - 1))
    {
        dprintf("%s --> %s\n", mailhost, buf);
        if (!isdigit(buf[0]) || buf[0] > '3')
        {
            logmsg("unexpected reply: %s", buf);
            exit(17);
        }

        if (buf[4] != '-')
            break;
    }

}


/*
**  say something to server and check the response
*/
void            chat(char *fmt,...)
{

    va_list         ap;
    char            buf[4096];

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    va_end(ap);

    sockprintf("%s", buf);
    sockflush();


    va_start(ap, fmt);
    dprintf("%s <-- ", mailhost);
    dvprintf(fmt, ap);
    va_end(ap);

    get_response();

}




void            bas64encode(FILE * infile)
{

    long            c,
                    qpos = 0,
                    triplet = 0,
                    lpos = 0;
    long            mask_table[] = {0xFC0000, 0x3F000, 0xFC0, 0x3F,};
    char            opos = 0;

    while ((c = fgetc(infile)) != EOF)
    {
        triplet |= (c << ((2 - qpos) * 8));
        qpos++;

        if (qpos == 3)
        {
            opos = 0;

            while (opos < 4)
            {
                sockputc(base64set[(triplet & mask_table[opos]) >> ((3 - opos) * 6)]);
                lpos++;
                opos++;
            }

            qpos = 0;
            triplet = 0;
        }

        if (lpos > 60)
        {
            sockprintf("\r\n");
            lpos = 0;
        }
    }

    if (qpos)
    {
        opos = 0;


        while (opos < (4 - (3 - qpos)))
        {
            sockputc(base64set[(triplet & mask_table[opos]) >> ((3 - opos) * 6)]);
            opos++;
        }
        for (; qpos < 3; qpos++)
            sockputc('=');
    }

    sockprintf("\r\n\r\n");

    sockflush();

}



unsigned long long
getmstime(void)
{

	struct timeval tv;

	if (gettimeofday(&tv, NULL) != 0)
		return (0);

	return (1000 * (unsigned long long) tv.tv_sec + (unsigned long long) tv.tv_usec / 1000 );

}



int
get_time(struct tm * tmLocal, int * iDiffHours, int * iDiffMins)
{

	int	iSecsDiff,
		iSignDiff,
		iMinutes;
	time_t	tCurr = time(NULL),
			tLocal,
			tGM;
	struct tm	tmTimeLOC = *localtime(&tCurr),
				tmTimeGM = *gmtime(&tCurr);

	*tmLocal = tmTimeLOC;

	tmTimeLOC.tm_isdst = 0;
	tmTimeGM.tm_isdst = 0;

	tLocal = mktime(&tmTimeLOC);
	tGM = mktime(&tmTimeGM);

	iSecsDiff = (int) difftime(tLocal, tGM);
	iSignDiff = (iSecsDiff >= 0) ? +1: -1;
	iMinutes = ((iSecsDiff >= 0) ? iSecsDiff: -iSecsDiff) / 60;

	*iDiffMins = iMinutes % 60;
	*iDiffHours = iSignDiff * (iMinutes / 60);


	return (0);

}




int
get_time_str(char *pszTimeStr, int iStringSize)
{

    int	iDiffHours = 0,
		iDiffMins = 0;
    struct tm	tmTime;
    char	szDiffTime[128] = "";

    get_time(&tmTime, &iDiffHours, &iDiffMins);


    if (iDiffHours > 0)
        sprintf(szDiffTime, " +%02d%02d", iDiffHours, iDiffMins);
    else
        sprintf(szDiffTime, " -%02d%02d", -iDiffHours, iDiffMins);


    strftime(pszTimeStr, iStringSize - strlen(szDiffTime) - 1, "%a, %d %b %Y %H:%M:%S", &tmTime);

    strcat(pszTimeStr, szDiffTime);


    return (0);

}




/*
**  main procedure
*/

struct option   options[] = {
    {"subject", 1, NULL, 's'},
    {"from", 1, NULL, 'f'},
    {"replay-to", 1, NULL, 'r'},
    {"errors-to", 1, NULL, 'e'},
    {"carbon-copy", 1, NULL, 'c'},
    {"smtp-host", 1, NULL, 'h'},
    {"smtp-port", 1, NULL, 'p'},
    {"body", 1, NULL, 'b'},
    {"mime-encode", 0, NULL, 'M'},
    {"attach", 1, NULL, 'a'},
    {"delete", 0, NULL, 'D'},
    {"verbose", 0, NULL, 'v'},
    {"version", 0, NULL, 'V'},
    {"help", 0, NULL, 'h'}
};





int             main(int argc, char **argv)
{

    char            buf[BUFSIZ];
    char            my_name[BUFSIZ];
    struct sockaddr_in sin;
    struct hostent *hp;
    int             c;
    int             i;
    int             delbody = 0;
    char           *cp;
    char           *defaultuser;
    char           *bodyfile = NULL;
    FILE           *bfp;
    int             nattch = 0;
    char           *attchf[MAX_FATTACH];
    char            boundary[256];
    char            timestr[128];


    while ((c = getopt_long(argc, argv, ":s:f:r:e:b:c:a:S:P:DMLvVh", options, NULL)) != EOF)
    {
        switch (c)
        {
            case 's':
                strcpy(subject, optarg);
                break;
            case 'f':
                strcpy(from_addr, optarg);
                break;
            case 'r':
                strcpy(reply_addr, optarg);
                break;
            case 'e':
                strcpy(err_addr, optarg);
                break;
            case 'c':
                strcpy(cc_addr, optarg);
                break;
            case 'b':
                bodyfile = optarg;
                break;
            case 'a':
                if (nattch < MAX_FATTACH)
                    attchf[nattch++] = optarg;
                mime_style = 1;
                break;
            case 'S':
                strcpy(mailhost, optarg);
                break;
            case 'P':
                mailport = atoi(optarg);
                break;
            case 'M':
                mime_style = 1;
                break;
            case 'D':
                delbody = 1;
                break;
            case 'v':
                verbose = 1;
                break;
            case 'V':
                version();
                exit(0);
            case 'h':
                usage(argv[0]);
                exit(0);
            default:
                fprintf(stderr, "SMTP: invalid option `%c'\n", optopt);
                fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]);
                exit(1);
        }
    }

    if (argc == optind)
    {
        fprintf(stderr, "SMTP: wrong number of arguments\n");
        fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]);
        exit(2);
    }

    if (bodyfile == NULL)
    {
        fprintf(stderr, "body file not specified\n");
        exit(3);
    }

    gethostname(host_name, sizeof(host_name));

    sprintf(boundary, "%llu.%u.%s", getmstime(), getpid(), host_name);


    if ((bfp = fopen(bodyfile, "rt")) == NULL)
    {
        perror(bodyfile);
        exit(4);
    }

/* Determine SMTP server */
    if (strlen(mailhost) == 0)
    {
        if ((cp = getenv("SMTPSERVER")) != NULL)
            strcpy(mailhost, cp);
        else
            strcpy(mailhost, "localhost");
    }

    if ((defaultuser = getenv("DEFAULTUSER")) == NULL)
        defaultuser = "postmaster";

/* Find out my own host name for HELO; if possible, get the FQDN. */
    if (gethostname(my_name, sizeof(my_name) - 1) < 0)
    {
        logmsg("gethostname: %s", errorstr(errno));
        if (delbody)
            remove(bodyfile);
        exit(5);
    }

    if ((hp = gethostbyname(my_name)) == NULL)
    {
        logmsg("%s: unknown host\n", my_name);
        if (delbody)
            remove(bodyfile);
        exit(6);
    }
    strcpy(my_name, hp->h_name);

/* Determine from address. */
    if (strlen(from_addr) == 0)
        sprintf(from_addr, "%s@%s", defaultuser, my_name);

/* Connect to smtp daemon on mailhost. */
    if ((hp = gethostbyname(mailhost)) == NULL)
    {
        logmsg("%s: unknown host\n", mailhost);
        if (delbody)
            remove(bodyfile);
        exit(7);
    }
    if (hp->h_addrtype != AF_INET)
    {
        logmsg("unknown address family: %d", hp->h_addrtype);
        if (delbody)
            remove(bodyfile);
        exit(8);
    }
    memset((char *) &sin, 0, sizeof(sin));
    memcpy((char *) &sin.sin_addr, hp->h_addr, hp->h_length);
    sin.sin_family = hp->h_addrtype;
    sin.sin_port = htons((short int) mailport);
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        logmsg("socket: %s", errorstr(errno));
        if (delbody)
            remove(bodyfile);
        exit(9);
    }
    if (connect(sock, (struct sockaddr *) & sin, sizeof(sin)) < 0)
    {
        logmsg("connect: %s", errorstr(errno));
        if (delbody)
            remove(bodyfile);
        exit(10);
    }

/* Give out SMTP headers. */
    get_response();             /* banner */
    chat("HELO %s\r\n", my_name);
    chat("MAIL FROM: <%s>\r\n", from_addr);
    for (i = optind; i < argc; i++)
        chat("RCPT TO: <%s>\r\n", argv[i]);
    if (strlen(cc_addr) > 0)
        chat("RCPT TO: <%s>\r\n", cc_addr);
    chat("DATA\r\n");

/* Give out Message header. */
    sockprintf("From: %s\r\n", from_addr);
    if (strlen(subject) > 0)
        sockprintf("Subject: %s\r\n", subject);

    if (strlen(reply_addr) > 0)
        sockprintf("Reply-To: %s\r\n", reply_addr);
    if (strlen(err_addr) > 0)
        sockprintf("Errors-To: %s\r\n", err_addr);
    sockprintf("Sender: %s@%s\r\n", defaultuser, my_name);

    sockprintf("To: %s", argv[optind]);
    for (i = optind + 1; i < argc; i++)
        sockprintf(",%s", argv[i]);
    sockprintf("\r\n");
    if (strlen(cc_addr) > 0)
        sockprintf("Cc: %s\r\n", cc_addr);

    get_time_str(timestr, sizeof(timestr));
    sockprintf("Date: %s\r\n", timestr);

    sockprintf("XMailer: %s\r\n", SMTPclient_Hello);

    if (mime_style)
    {
        sockprintf(
                "MIME-Version: 1.0\r\n"
                "Content-Type: multipart/mixed;\r\n"
                "\tboundary=\"%s\"\r\n", boundary);
    }

    sockprintf("\r\n");

    if (mime_style)
    {
        sockprintf("This is a multi-part message in MIME format.\r\n"
                "\r\n"
                "--%s\r\n"
                "Content-Type: text/plain;\r\n"
                "\tcharset=\"iso-8859-1\"\r\n"
                "Content-Transfer-Encoding: 7bit\r\n\r\n", boundary);
    }

/* Give out Message body. */
    while (fgets(buf, sizeof(buf), bfp))
    {
        buf[strlen(buf) - 1] = 0;

        if (strcmp(buf, ".") == 0)
        {                       /* quote alone dots */
            sockprintf("..\r\n");
        }
        else
        {                       /* pass thru mode */
            sockprintf("%s\r\n", buf);
        }
    }

    fclose(bfp);

    for (--nattch; nattch >= 0; nattch--)
    {
        FILE           *afp = fopen(attchf[nattch], "rb");

        if (afp == NULL)
            perror(attchf[nattch]);
        else
        {
            char const *	slashptr = strrchr(attchf[nattch], '/');

            if (slashptr == NULL)
                slashptr = attchf[nattch];
            else
            	++slashptr;

            sockprintf("\r\n--%s\r\n"
                    "Content-Type: application/octet-stream;\r\n"
                    "\tname=\"%s\"\r\n"
                    "Content-Transfer-Encoding: base64\r\n"
                    "Content-Disposition: attachment;\r\n"
                    "\tfilename=\"%s\"\r\n\r\n", boundary, slashptr, slashptr);


            bas64encode(afp);


            fclose(afp);
        }
    }

    if (mime_style)
        sockprintf("--%s--\r\n", boundary);

    chat(".\r\n");
    chat("QUIT\r\n");

    close(sock);

    if (delbody)
        remove(bodyfile);

    exit(0);

}
